#include "XS_Base.H"

#include "Phase_Space_Handler.H"
#include "Running_AlphaS.H"
#include "Combined_Selector.H"
#include "Standard_Selector.H"
#include "Run_Parameter.H"
#include "Message.H"

#include <stdio.h>

using namespace EXTRAXS;

XS_Base::XS_Base()
{
  m_name="Empty XS";
} 

XS_Base::XS_Base(const size_t nin,const size_t nout,const ATOOLS::Flavour *flavours,
		 const int scalescheme,const int kfactorscheme,const double scalefactor,
		 BEAM::Beam_Spectra_Handler *const beamhandler,PDF::ISR_Handler *const isrhandler,
		 ATOOLS::Selector_Data *const selectordata):
  Integrable_Base(nin,nout,flavours,scalescheme,kfactorscheme,scalefactor,
		  beamhandler,isrhandler,selectordata),
  p_colours(NULL)
{
  Init(flavours);
  ResetSelector(selectordata);
  p_pshandler = new PHASIC::Phase_Space_Handler(this,isrhandler,beamhandler);
}

XS_Base::XS_Base(const size_t nin,const size_t nout,const ATOOLS::Flavour *flavours):
  Integrable_Base(nin,nout,flavours),
  p_colours(NULL)
{
  Init(flavours);
  p_selector = new ATOOLS::No_Selector();
}

XS_Base::~XS_Base() 
{
 if (p_colours!=NULL) { 
    for (size_t i=0;i<m_nin+m_nout;++i) delete p_colours[i];
    delete [] p_colours;
  }
}

void XS_Base::Init(const ATOOLS::Flavour *flavours)
{
  p_flavours = new ATOOLS::Flavour[m_nin+m_nout];
  if (flavours!=NULL) {
    for (size_t i=0;i<m_nin+m_nout;++i) p_flavours[i]=flavours[i];
    m_name=GenerateName(m_nin,m_nout,flavours);
  }
  p_colours = new int*[m_nin+m_nout];
  for (size_t i=0;i<m_nin+m_nout;++i) { 
    p_colours[i] = new int[2]; 
    p_colours[i][0]=p_colours[i][1]=0; 
  }  
  double massin=0.0, massout=0.0;
  for (size_t i=0;i<m_nin;++i) massin+=p_flavours[i].Mass();
  for (size_t i=m_nin;i<m_nout;++i) massout+=p_flavours[i].Mass();
  if (massin>massout) m_threshold=ATOOLS::sqr(massin);
  else m_threshold=ATOOLS::sqr(massout);
}

std::string XS_Base::GenerateName(const size_t nin,const size_t nout,
				  const ATOOLS::Flavour *flavours) 
{
  char help[20];
  std::string name;
  sprintf(help,"%i",nin);
  name=std::string(help);
  name+=std::string("_");
  sprintf(help,"%i",nout);
  name+=std::string(help);
  name+=std::string("__");
  for (size_t i=0;i<nin+nout;) {
    name+=std::string(flavours[i].Name());
    if (flavours[i].IsAnti()) {
      if (flavours[i].Kfcode()==ATOOLS::kf::e ||
	  flavours[i].Kfcode()==ATOOLS::kf::mu ||
	  flavours[i].Kfcode()==ATOOLS::kf::tau ||
	  flavours[i].Kfcode()==ATOOLS::kf::Hmin) {
	name.replace(name.length()-1,1,"+");
      }
      else {
	name+=std::string("b"); 
      }
    }
    name+=std::string("_");
    if (++i==nin) name+=std::string("_");
  }
  return name.substr(0,name.length()-1);
}

double XS_Base::Differential(const ATOOLS::Vec4D *momenta) 
{
  SetMomenta(momenta);
  SetSTU(momenta);
  return Differential(m_s,m_t,m_u);
}

bool XS_Base::SetColours(const ATOOLS::Vec4D *momenta) 
{
  SetMomenta(momenta);
  SetSTU(momenta);
  SetScale(momenta[2].PPerp2());  
  return SetColours(m_s,m_t,m_u);
}

double XS_Base::Scale(const ATOOLS::Vec4D *momenta) 
{
  SetMomenta(momenta);
  if (m_nin==1) return momenta[0].Abs2();
  SetSTU(momenta);
  double pt2;
  double MZ=ATOOLS::sqr(ATOOLS::Flavour(ATOOLS::kf::Z).Mass());
  switch (m_scalescheme) {
  case 1:
    if (m_nin+m_nout==4) m_scale=momenta[2].PPerp2();
    return m_scale;
  case 2:
    return m_scale;
  case 10:
    m_scale=(momenta[0]+momenta[1]).PPerp2();
    return m_scale=pow(pow(m_scale,6.)*pow(MZ,1.),2./(12.+1.));
  case 11:
    pt2=2.*m_s*m_t*m_u/(m_s*m_s+m_t*m_t+m_u*m_u);
    return m_scale=pt2;
  case 12:
    return m_scale=(momenta[0]+momenta[1]).PPerp2();
  case 20:
    return 1.;
  default:
    return m_scale=m_s;
  }
}

double XS_Base::KFactor(const double scale) 
{
  switch (m_kfactorscheme) {
  case 1:
    return pow(MODEL::as->AlphaS(scale*m_scalefactor)/
	       MODEL::as->AlphaS(ATOOLS::sqr(ATOOLS::rpa.gen.Ecms())),m_nin+m_nout-2);
  default:
    return 1.;
  }
}

void XS_Base::SwapInOrder() 
{
  std::swap(p_flavours[0],p_flavours[1]);
  std::swap(p_momenta[0],p_momenta[1]);
  std::swap(p_colours[0],p_colours[1]);
  m_swaped=true;
}

void XS_Base::RestoreInOrder() 
{
  if (m_swaped) {
    std::swap(p_flavours[0],p_flavours[1]);
    std::swap(p_momenta[0],p_momenta[1]);
    std::swap(p_colours[0],p_colours[1]);
    m_swaped=false;
  }
}

void XS_Base::ResetSelector(ATOOLS::Selector_Data *const selectordata)
{
  if (p_selector!=NULL) delete p_selector;
  if (selectordata!=NULL) {
    p_selector = new ATOOLS::Combined_Selector(m_nin,m_nout,p_flavours,selectordata);
  }
  else {
    ATOOLS::msg.Error()<<"XS_Base::ResetSelector("<<selectordata<<"): "
		       <<"(\""<<m_name<<"\")"<<std::endl
		       <<"   No cuts specified. Initialize 'No_Selector'."<<std::endl;
    p_selector = new ATOOLS::No_Selector();
  }
}

void XS_Base::SetSTU(const ATOOLS::Vec4D *momenta)
{
  m_s=(momenta[0]+momenta[1]).Abs2();
  m_t=(momenta[0]-momenta[2]).Abs2();
  m_u=(momenta[0]-momenta[3]).Abs2();
}

void XS_Base::SetMax(const double max,const int flag)             
{ 
  SetMax();
}

void XS_Base::SetMax()             
{ 
}

XS_Base *const XS_Base::operator[](const size_t i) const 
{
  return NULL;
}

size_t XS_Base::Size() const
{ 
  return 0; 
}

void XS_Base::SelectOne()
{ 
}

void XS_Base::DeSelect()
{ 
}

